// First price auction with private values
// BAB 11-3-2012
// Updated 11-6-2013

#include "fpaknown.hpp"

void solveFPA(int nvals, int nbids, 
	      double entryCost, double reservePrice);

int main(int argc, char ** argv)
{
  double entryCost=0.0;
  double reservePrice=0.0;
  int nvals=2;
  int nbids=50;

  solveFPA(nvals,nbids,entryCost,reservePrice);
  return 0;
}

void solveFPA(int nvals, int nbids, 
	      double entryCost, double reservePrice)
{
  int t, a, ahat;
  double minAngleIncrement = 0.05;
  vector< vector<double> > devObjectives;

  bool verbose = true;
  
  if (verbose)
    cout << "Starting main" << endl;

  try
    {
      FPAKnown fpa(nbids,nvals,entryCost,reservePrice);
      fpa.highbid=1.0;
      fpa.distribution.clear();
      fpa.distribution.push_back(vToTheAlpha(1.0),1.0); // uniform
      // fpa.distribution.push_back(independent(1.0/3.0),1.0);
      // fpa.distribution.push_back(uniformNoTie(nvals),1.0);
      // fpa.distribution.push_back(independent(1.0/3.0),1.0);

      BCESolver solver(fpa);

      solver.setParameter(BCESolver::MinAngleIncrement,minAngleIncrement);
      solver.setParameter(BCESolver::OnlyICDown,false);
      solver.setParameter(BCESolver::OnlyICUp,false);
      solver.setParameter(BCESolver::DisplayLevel,1);

      if (verbose)
	cout << "Constructor finished" << endl;

      solver.populate();
      if (verbose)
	cout << "Done populating" << endl;

      char filename[100];

      sprintf(filename,"fpaknown_nv=%d_nb=%d.bce",
      	      nvals,nbids);

      IloCplex cplex = solver.getCplex();

      cplex.setParam(IloCplex::RootAlg,IloCplex::Barrier);
      cplex.setParam(IloCplex::SimDisplay,0);
      solver.setParameter(BCESolver::DisplayLevel,1);

      // Maximize total surplus
      cplex.getObjective().setSense(IloObjective::Maximize);
      cplex.getObjective().setExpr(solver.getObjectiveFunction(3));      

      if (verbose)
      	cout << "Objective function set" << endl;
      solver.solve();

      cout << cplex.getValue(solver.getObjectiveFunction(4)) << endl;
      cout << cplex.getValue(solver.getObjectiveFunction(5)) << endl;

      cout << "Total surplus: " << cplex.getValue(solver.getObjectiveFunction(3)) << endl;

      // Add a constraint that equilibrium must be efficient and
      // minimize revenue
      cout << "Adding constraint" << endl;
      cplex.getModel().add(solver.getObjectiveFunction(3) == cplex.getObjValue());

      cplex.getObjective().setExpr(-1*solver.getObjectiveFunction(2));
      solver.solve();

      // Print out some results
      cout << "Objective = " << setprecision(16) << cplex.getObjValue() << endl;
      cout << "Player 0's payoff = " << cplex.getValue(solver.getObjectiveFunction(0)) << endl;
      cout << "Player 1's payoff = " << cplex.getValue(solver.getObjectiveFunction(1)) << endl;
      cout << "Revenue = " << cplex.getValue(solver.getObjectiveFunction(2)) << endl;

      cout << cplex.getValue(solver.getObjectiveFunction(0)) << ";..." << endl;


      // Now map the efficient frontier of bidders' surpluses
      solver.mapBoundary();


      // Save the computations
      BCEData data;
      solver.getData(data);
      data.setNumValues(vector<int>(2,nvals));

      if (verbose)
      	cout << "Saving data..." << endl;

      BCEData::save(data,filename);

    }
  catch (IloException & e)
    {
      cerr << "Concert exception caught: " << e << endl
	   << e.getMessage() << endl;
    }
  catch (BCEException & err)
    {
      cerr << "BCEException caught." << endl;
    }
  catch (...)
    {
      cout << "Unknown exception caught." << endl;
    }

}
